{#########################################################################}
{#                                                                       #}
{#                          Programm Linie_3d                            #}
{#                                                                       #}
{# Das Programm dient zur Ansteuerung der Faesmaschine.                  #}
{#                                                                       #}
{#                                                                       #}
{#-----------------------------------------------------------------------#}
{#         Programmiersprache........: Turbo Pascal 6.0                  #}
{#         Rechnertyp ...............: 486                               #}
{#         Letzte nderung...........: 08.03.97                          #}
{#         Grafikkarte...............: VGA                               #}
{#                                                                       #}
{#         geschrieben von ..........: Clemens Simon                     #}
{#         Kontaktadresse............: Giselastr.6, D-82319 Starnberg    #}
{#         Telefon ..................: 08151/3838                        #}
{#                                                                       #}
{#########################################################################}
program Bresenham_3d;
uses dos,Graph, crt;
type xyz_i = record
         x : integer;
         y : integer;
         z : integer;
         end;
     xyz   = record
         x : real;
         y : real;
         z : real;
         end;
     xyz_b  = record
         x : boolean;
         y : boolean;
         z : boolean;
         end;
     achstype= record
         takt         : byte;   { Taktbyte der Steuerung fuer Takt}
         richtung     : byte;   { Richtungbyte der Steuerung fuer Achse }
         text         : char;   { Namen der Achse  z.B. x,y,z,..}
         endschalter  : byte;   { Endschalter Bit der Steuerung }
         max_schritte : integer;{ Maximaler Verfahrweg in Schritten }
         min_schritte : integer;{ Minimaler Verfahrweg in Schritten }
         steigung     : integer;{ Spindelsteigung der Achse }
         Schritte_rpm : integer;{ Anzahl der Schritte pro Umdrehung }
         maxzeit      : integer;{ Maximale Zeit fr Bewegung einer Achse in ms}
         skal_faktor  : integer;{ Skalierungsfaktor der Achse }
         endstop      : boolean;{ true wenn Endschalter angefahren ist}
         altbit       : byte   ;{ Letzter Pegel der Achse}
         end;

           {Konstanten}
           {**********}

const BasisAdrDio            = $378;  {Basisadresse Digital-Interface}
      Kontroll               = 145 ;  {Kontrollsignal=Port A=Eingang,B=Ausgang,C untere }
                                      {4 Bit auf Eingabe,C obere 4 Bit auf Ausgabe.}
      achszahl               = 3;     {Anzahl der Achsen}

var     sollwert             : array[1..2] of xyz;
        weg                  : xyz;
        fraesweg             : array[1..2] of xyz;
        schritte             : array[1..2] of xyz_i;
        offset_fraese        : xyz;
        schritte_rpm         : xyz_i;
        istposition          : xyz_i; {Istposition der NC-Maschine in Schritten}
        grDriver             : Integer;
        grMode               : Integer;
        ErrCode              : Integer;
        offset               : array[1..3] of xyz_i;
        i                    : integer;
        wartezeit            : integer;
        achse                : array[1..achszahl] of achstype;

{----------------------------------------------------------------------------}
{             Systemzeit in Hundertstel Sekunden holen                       }
{----------------------------------------------------------------------------}
function zeit:integer;
var
  h, m, s, hund : Word;
begin
  GetTime(h,m,s,hund);
  zeit:=((h*24+m)*60+s)*100+hund;
end;

{----------------------------------------------------------------------------}
function sgn_01 (wert:real):real;
{----------------------------------------------------------------------------}
{ SGN_01  Signum Funktion liefert das Vorzeichen einer Zahl als Ergebnis
  zurck. Der Wert ist 1 bei positivem Vorzeichen und 0 bei negativem
  Vorzeichen}
begin
     if wert<0 then sgn_01:=0 else sgn_01:=1;
end;

{----------------------------------------------------------------------------}
function sgn (wert:integer):integer;
{----------------------------------------------------------------------------}
{ SGN  Signum Funktion liefert das Vorzeichen einer Zahl als Ergebnis
  zurck. Der Wert ist 1 bei positivem Vorzeichen und -1 bei negativem
  Vorzeichen}
begin
     if wert<0 then sgn:=-1 else sgn:=1;
end;

{----------------------------------------------------------------------------}
           {Endschalter abfragen und anzeigen (Rechneransteuerung)}
{----------------------------------------------------------------------------}

function Endschalter(j:integer):boolean;
var hilfsvar: byte;
begin
   {Annahme: alle Endschalter sind nicht angefahren }
   achse[1].endstop:=false;
   achse[2].endstop:=false;
   achse[3].endstop:=false;
   {Vorbelegung: Endschalter ist angefahren und wird, falls nicht
                   angefahren, weiter unten zurueckgesetzt.}
   endschalter:=true;

   hilfsvar:=port [BasisAdrDio +2];
   hilfsvar:=(hilfsvar mod 5);  {Kappt die letzten 3 Bit}
   case hilfsvar of
   4: achse[1].endstop:=true;   {Endschalter der x-Achse}
   3: achse[2].endstop:=true;   {Endschalter der y-Achse}
   1: achse[3].endstop:=true;   {Endschalter der z-Achse}
   else
   {Kein Endschalter angefahren ansonsten angefahren wie Vorbelegung!}
   endschalter:=false;
   end;
end;

{----------------------------------------------------------------------------}
procedure koord_achsen_zeichen;
{----------------------------------------------------------------------------}
          {Koordinatenachsen zeichnen}
begin
          setcolor(2); {grn}
          {Vorderansicht}
          line(offset[1].x,offset[1].y,offset[1].x+100,offset[1].y);
          line(offset[1].x,offset[1].y,offset[1].x,offset[1].y-100);
          {Seitenansicht}
          line(offset[2].x,offset[2].y,offset[2].x-100,offset[2].y);
          line(offset[2].x,offset[2].y,offset[2].x,offset[2].y-100);
          {Draufsicht}
          line(offset[3].x,offset[3].y,offset[3].x+100,offset[3].y);
          line(offset[3].x,offset[3].y,offset[3].x,offset[3].y-100);
          {Koordinatenachsen beschriften}
          outtextxy(offset[1].x+100,offset[1].y    ,'x');
          outtextxy(offset[1].x    ,offset[1].y-100,'z');
          outtextxy(offset[1].x    ,offset[1].y    ,'y');

          outtextxy(offset[2].x-110,offset[2].y    ,'y');
          outtextxy(offset[2].x    ,offset[2].y-100,'z');
          outtextxy(offset[2].x    ,offset[2].y    ,'x');

          outtextxy(offset[3].x+100,offset[3].y    ,'x');
          outtextxy(offset[3].x    ,offset[3].y-100,'y');
          outtextxy(offset[3].x    ,offset[3].y    ,'z');
          setcolor(7); {wei}
end;



{----------------------------------------------------------------------------}
procedure  fraesweg_in_schritte(fraesweg    :xyz;
                                var schritte:xyz_i);
{----------------------------------------------------------------------------}

{Rechnet den Weg aus, den eine Maschinenachse in Schritten durchfhren mu}
begin
         schritte.x:=round(fraesweg.x*achse[1].schritte_rpm/achse[1].steigung);
         schritte.y:=round(fraesweg.y*achse[2].schritte_rpm/achse[2].steigung);
         schritte.z:=round(fraesweg.z*achse[3].schritte_rpm/achse[3].steigung);
end;

{----------------------------------------------------------------------------}
procedure  schritte_in_fraesweg(schritte:xyz_i;
                                var fraesweg:xyz);
{----------------------------------------------------------------------------}

{Rechnet den Weg aus, den eine Maschinenachse in Schritten durchfhrt hat}
begin
         fraesweg.x:=schritte.x/achse[1].schritte_rpm*achse[1].steigung/achse[1].skal_faktor;
         fraesweg.y:=schritte.y/achse[2].schritte_rpm*achse[2].steigung/achse[2].skal_faktor;
         fraesweg.z:=schritte.z/achse[3].schritte_rpm*achse[3].steigung/achse[3].skal_faktor;
end;


{----------------------------------------------------------------------------}
procedure koord_real_fraese(sollwert    : xyz;
                           offset      : xyz;
                      var  weg         :xyz);
{----------------------------------------------------------------------------}
(* Umrechnung der mm-Angaben in das Maschinenkoordinatensystem *)
begin
   weg.x:= achse[1].skal_faktor*sollwert.x+offset.x;
   weg.y:= achse[2].skal_faktor*sollwert.y+offset.y;
   weg.z:= achse[3].skal_faktor*sollwert.z+offset.z;
end;


{----------------------------------------------------------------------------}
procedure NC_Initialisierung;
{----------------------------------------------------------------------------}
begin
   port [BasisAdrDio +3] := Kontroll;
   DELAY(100);
end;


{----------------------------------------------------------------------------}
procedure warten(istschritt:integer;
                 Maxschritt:integer;
                 speed     :integer);
{----------------------------------------------------------------------------}

const      pi = 3.141592;
       faktor = 2;
       minzeit= 1;
var
            i : integer;
    geschwindigkeit : Integer;
begin
    geschwindigkeit:=round((100-speed)/10);
    if geschwindigkeit<0 then geschwindigkeit:=0;
    if geschwindigkeit>100 then geschwindigkeit:=100;
    {Kurz Bechleunigungs und Bremsrampen}
    if maxschritt<200 then
    begin
       wartezeit:=round(abs(cos(istschritt/maxschritt*pi)+1)*faktor)+minzeit+geschwindigkeit;
       {putpixel(istschritt,wartezeit,4);}
       delay(wartezeit);
    end
    else
    begin
       {Beschleunigungsrampe}
       if istschritt<100 then
       begin
          wartezeit:=round(abs(cos(istschritt/100*pi)+1)*faktor)+minzeit+geschwindigkeit;
          delay(wartezeit);
       end
       else
       {Bremsrampe}
       if maxschritt-istschritt<100 then
       begin
          wartezeit:=round(abs(cos((maxschritt-istschritt)/100*pi)+1)*faktor)+minzeit+geschwindigkeit;
          delay(wartezeit);
       end
       else
       {Fahren mit vorgegebener Geschwindigkeit}
       delay(minzeit);

    end;
end;



{----------------------------------------------------------------------------}
procedure NC_Linie_relativ(P1         :xyz_i;
                           p2         :xyz_i;
                           speed      :integer;
                           endstop    :byte);
{----------------------------------------------------------------------------}

var
        d                    : xyz_i;
        d_schritt            : xyz;
        p_aktuell            : xyz_i;
        richtung             : xyz_i;
        richtungsbyte        : word;
        bit                  : xyz_i;
        ausgabe_byte         : word;
        Max_anzahl_schritte  : integer;
        i                    : integer;
        schalter             : integer;
        pos_richtung         : xyz_i;
label   fahrenabbruch;
begin
          {Bestimmen der Steigungen der Geraden im Raum}
          d.x:=P2.x-P1.x;
          d.y:=P2.y-P1.y;
          d.z:=P2.z-P1.z;

          {Bestimmen der Achse, die am meisten Schritte fahren mu}
          Max_anzahl_schritte:=abs(d.x);
          {Prfen, ob y-Richtung grssere Verfahrlnge ist}
          if abs(d.y)>Max_anzahl_schritte then  {..ja}
          begin
             Max_anzahl_schritte:=abs(d.y);
          end; {if}
          {Prfen, ob z-Richtung grssere Verfahrlnge ist}
          if abs(d.z)>Max_anzahl_schritte then
          begin
             Max_anzahl_schritte:=abs(d.z);
          end; {if}
          {Nur Verfahren, wenn berhaupt Max_anzahl_schritte >0}
          if Max_anzahl_schritte<>0 then
          begin
             {Steigung auf Schrittzahl normieren}
             d_schritt.x:=abs(d.x/Max_anzahl_schritte);
             d_schritt.y:=abs(d.y/Max_anzahl_schritte);
             d_schritt.z:=abs(d.z/Max_anzahl_schritte);
             {Richtung als Binaerwert bestimmen}
             richtung.x:=round(sgn_01(d.x)*achse[1].richtung);
             richtung.y:=round(sgn_01(d.y)*achse[2].richtung);
             richtung.z:=round(sgn_01(d.z)*achse[3].richtung);
             {Zusammensetzen des Richtungsbytes}
             richtungsbyte:=richtung.x+richtung.y+richtung.z;

             {Vorzeichen fr Positionsbestimmung ermitteln}
             pos_richtung.x:=sgn(d.x);
             pos_richtung.y:=sgn(d.y);
             pos_richtung.z:=sgn(d.z);

             {Linienalgorithmus: Bestimmt die Ausgabewerte in Byteform zur
             Steuerung der NC-Maschine}
             for i:=0 to Max_anzahl_schritte do
             begin
               {Ein Bit wird verndert, wenn sich i*d_schritt erhht}
               {Bei geraden Zahlen i*d_schritt wird eine 0
                bei ungeraden Zahlen i*d_schritt wird eine 1 fr
                bit.x gesetzt}
               bit.x:=((round(i*d_schritt.x+achse[1].altbit))mod 2);
               bit.y:=((round(i*d_schritt.y+achse[2].altbit))mod 2);
               bit.z:=((round(i*d_schritt.z+achse[3].altbit))mod 2);
               {Ausgangsbyte wird zusammengesetzt und mit den Bits der
               kanle multipliziert}
               ausgabe_byte:=bit.x*achse[1].takt+bit.y*achse[2].takt+bit.z*achse[3].takt+richtungsbyte;
               {Ausgabeport ansteuern}
               port [BasisAdrDio +1] := ausgabe_byte;
               {Prfen, ob Enschalter angefahren wurde}
               {Nur bercksichtigen wenn endstop nicht 0 ist}
               if endstop=0 then
                  if (endschalter(schalter) or keypressed) then goto fahrenabbruch;
               {Pause einlegen}
               warten(i,max_anzahl_schritte,speed);
             end; {for}
             {Schrittstellung speichern}
             achse[1].altbit:=bit.x;
             achse[2].altbit:=bit.y;
             achse[3].altbit:=bit.z;
         end;
   fahrenabbruch:
   {Istposition mitspeichern}
   istposition.x:=round(d_schritt.x*i)+istposition.x;
   istposition.y:=round(d_schritt.y*i)+istposition.y;
   istposition.z:=round(d_schritt.z*i)+istposition.z;

end;

{----------------------------------------------------------------------------}
procedure NC_Nullpunkt_anfahren(achsezahl:byte);
{----------------------------------------------------------------------------}

var     j                : integer;
        schalter         : xyz_b;
        aktuelle_achse   : word;
        takt             : byte;
        timer            : integer;
        hilfsvar         : boolean;
        startzeit        : integer;
        richtung         : byte;
        ausgabe          : word;
        fehlertext       : string;
        fehler           : boolean;
        zeitstr          : string;
        p1               : xyz_i;
        p2               : xyz_i;

label             Abbruch;
begin
   writeln('Endschalter simulieren durch drcken beliebiger Taste');
   for j:=1 to achszahl do
   begin
      timer:=0;
      writeln ('Nullstellung der ', achse[j].text, '- Achse wird angefahren ');
      takt:=1;
      startzeit:=zeit;  {Startzeit holen}
      {Die Richtung ist immer negativ, da zum Endschalter gefahren werden soll}
      richtung:=round(sgn_01(-achse[j].skal_faktor));
      while not(achse[j].endstop or (timer>achse[j].maxzeit) or keypressed) do
      begin {Verfahren der aktuellen Achse bis zum Endschalter}
         {Endschalter abfragen}
         endschalter(j);
         inc(takt);
         takt:=takt mod 2; {Takt fuer aktuelle Achse bestimmen}
         ausgabe:=richtung*achse[j].richtung+achse[j].takt*takt;
         port [BasisAdrDio +1] := ausgabe;                  {Ausgeben des Bits}
         delay(1);
         timer:=zeit-startzeit;
      end; {while}
      if timer > achse[j].maxzeit then
      begin { berschreitung der maximalen Verfahrzeit; Timeout}
         fehlertext:=achse[j].text+'-Achse konnte innerhalb von ';
         str(round(achse[j].maxzeit/1000),zeitstr);
         fehlertext:=fehlertext+zeitstr+'Sekunden Nullpunkt nicht erreichen';
         fehler:= true;
         writeln(fehlertext);
      end
      else  fehler:=false;
      {Prfen, ob Endschalter von aktueller Achse oder von einer anderen Achse ausgelst wurde}
      if achse[j].endstop=false then {Falscher Endschalter angefahren}
      begin
         fehlertext:='Endschalter der falschen Achse angefahren.';
         fehlertext:=fehlertext+' Prfen Sie, ob die Kabel richtig eingesteckt sind';
         fehlertext:=fehlertext+' , bzw. ob die Adressen der Enschalter stimmen';
         fehler:=true;
         writeln(fehlertext);
      end
      else fehler:=false;
      {Tastaturpuffer leeren}
      while keypressed do readkey;
   end;{for}
  { Vom Endschalter fahren}
  p1.x:=0;
  p1.y:=0;
  p1.z:=0;
  p2.x:=1000;
  p2.y:=1000;
  p2.z:=1000;
  {Auf Nullpunkt (Weltkoordinatensystem) fahren}
  write('Grundstellung anfahren..........');
  NC_Linie_relativ(p1,p2,90,1);
  writeln(' erreicht.');
  {Zaehler auf Null stellen}
  istposition.x:=0;
  istposition.y:=0;
  istposition.z:=0;
  readln;

end;

{----------------------------------------------------------------------------}
procedure NC_AutomatikEin;
{----------------------------------------------------------------------------}

var
   out : byte;

begin
   out := 128;                                    {Umschaltbit ist 128}
   port [BasisAdrDio +2] := out;                  {Ausgeben des Bits}
end;



{----------------------------------------------------------------------------}
procedure NC_Reset;{ Fhrt einen Reset fr den angewhlten Port aus}
{----------------------------------------------------------------------------}

var reset:word;
begin
  reset := 0;
  port [BasisAdrDio +1] := reset;
  delay(100);
  port [BasisAdrDio +2] := reset;
  delay(100);
end;

{===========================================================================}
{                            Hauptprogramm                                  }
{===========================================================================}
begin
   {Urspungspunkte der Koordinatensysteme in Bildschirmkoordinaten}
   (*Vorderansicht*)
   offset[1].x:=100;
   offset[1].y:=150;
   (*seitenansicht*)
   offset[2].x:=500;
   offset[2].y:=150;
   (*Draufsicht   *)
   offset[3].x:=100;
   offset[3].y:=400;
   {Parameter der Fraesmaschine}
   offset_fraese.x:=0;
   offset_fraese.y:=0;
   offset_fraese.z:=0;
   achse[1].takt        := 1  ;   { Taktbyte der Steuerung fuer Takt x-Achse}
   achse[2].takt        := 2  ;   { Taktbyte der Steuerung fuer Takt y-Achse}
   achse[3].takt        := 4  ;   { Taktbyte der Steuerung fuer Takt z-Achse}
   achse[1].richtung    := 8  ;   { Richtungbyte der Steuerung fuer x-Achse }
   achse[2].richtung    := 16 ;   { Richtungbyte der Steuerung fuer y-Achse }
   achse[3].richtung    := 32 ;   { Richtungbyte der Steuerung fuer z-Achse }
   achse[1].text        := 'x';   { Namen der x-Achse             }
   achse[2].text        := 'y';   { Namen der y-Achse             }
   achse[3].text        := 'z';   { Namen der z-Achse             }
   achse[1].endschalter := 1  ;   { Endschalter Bit der Steuerung fuer x-Achse}
   achse[2].endschalter := 2  ;   { Endschalter Bit der Steuerung fuer x-Achse}
   achse[3].endschalter := 4  ;   { Endschalter Bit der Steuerung fuer x-Achse}
   achse[1].max_schritte:= 20700; { Maximaler Verfahrweg in Schritten der x-Achse}
   achse[2].max_schritte:= 24000; { Maximaler Verfahrweg in Schritten der x-Achse}
   achse[3].max_schritte:= 2000;  { Maximaler Verfahrweg in Schritten der x-Achse}
   achse[1].min_schritte:= 0;     { Minimaler Verfahrweg in Schritten der x-Achse}
   achse[2].min_schritte:= 0;     { Minimaler Verfahrweg in Schritten der x-Achse}
   achse[3].min_schritte:= 0;     { Minimaler Verfahrweg in Schritten der x-Achse}
   achse[1].steigung    := 4;     { Spindelsteigung der x-Achse }
   achse[2].steigung    := 4;     { Spindelsteigung der y-Achse }
   achse[3].steigung    := 5;     { Spindelsteigung der z-Achse }
   achse[1].Schritte_rpm:= 200;   { Anzahl der Schritte pro Umdrehung der x-Achse}
   achse[2].Schritte_rpm:= 200;   { Anzahl der Schritte pro Umdrehung der x-Achse}
   achse[3].Schritte_rpm:= 200;   { Anzahl der Schritte pro Umdrehung der x-Achse}
   achse[1].maxzeit     := 10000; { maximale Verfahrzeit bevor timeout }

   achse[2].maxzeit     := 5000;  { maximale Verfahrzeit bevor timeout }
   achse[3].maxzeit     := 1000;  { maximale Verfahrzeit bevor timeout }
   achse[1].skal_faktor := 2  ;  { Skalierungsfaktor fr x-Achse }
   achse[2].skal_faktor := 2  ;  { Skalierungsfaktor fr y-Achse }
   achse[3].skal_faktor := 2  ;  { Skalierungsfaktor fr z-Achse }


   clrscr;
   {Initialisierung des Ausgangsports}
   NC_reset;
   NC_Initialisierung;
   {Umschalten der NC-Steuerung von Hand auf Automatikbetrieb}
   NC_AutomatikEin;
NC_Nullpunkt_anfahren(achszahl);
   {Grafik initialisieren}
   grDriver := Detect;
   writeln(grdriver);
   {in der nchsten Zeile bitte Pfad zum BGI Verzeichnis eingeben}
   InitGraph(grdriver, grMode,'c:\bp\bgi');
   ErrCode := GraphResult;
   if ErrCode = grOk then
   begin  { Do graphics }
          {erster Punkt der Linie in mm}
          sollwert[1].x:=0;
          sollwert[1].y:=0;
          sollwert[1].z:=0;
          {zweiter Punkt der Linie in mm}
          sollwert[2].x:=6;
          sollwert[2].y:=3;
          sollwert[2].z:=0;
          koord_achsen_zeichen;
          for i:=1 to 2 do
          begin
             (* Umrechnung der mm-Angaben in das Maschinenkoordinatensstem *)
             koord_real_fraese(sollwert[i],offset_fraese,fraesweg[i]);
             (* Umrechnung der mm- Angaben in Schrittzahlen *)
             fraesweg_in_schritte(fraesweg[i],schritte[i]);
          end; {for}
          NC_Linie_relativ(schritte[1],schritte[2],100,0);

(*               {Neuen Punkt fuer Bildschirmgrafik berechnen}
               p_aktuell.x:=round((schritte[1].x+round(d_schritt.x*i))/10);
               p_aktuell.y:=round((schritte[1].y+round(d_schritt.y*i))/10);
               p_aktuell.z:=round((schritte[1].z+round(d_schritt.z*i))/10);
               {Punkte der Linie in den 3 Koordinatensystemen setzen}
               putpixel( P_aktuell.x+offset[1].x,-P_aktuell.z+offset[1].y,7);
               putpixel(-P_aktuell.y+offset[2].x,-P_aktuell.z+offset[2].y,7);
               putpixel( P_aktuell.x+offset[3].x,-P_aktuell.y+offset[3].y,7);


          {Punkte P1 und P2 beschriften}
          {Vorderansicht}
          outtextxy( schritte[1].x+offset[1].x,-schritte[1].z+offset[1].y,'P1');
          outtextxy(-schritte[1].y+offset[2].x,-schritte[1].z+offset[2].y,'P1');
          {Seitenansicht}
          outtextxy( schritte[1].x+offset[3].x,-schritte[1].y+offset[3].y,'P1');
          outtextxy( schritte[2].x+offset[1].x,-schritte[2].z+offset[1].y,'P2');
          {Draufsicht}
          outtextxy(-schritte[2].y+offset[2].x,-schritte[2].z+offset[2].y,'P2');
          outtextxy( schritte[2].x+offset[3].x,-schritte[2].y+offset[3].y,'P2');
*)
          readln;
          CloseGraph;
        schritte_in_fraesweg(istposition,weg);
          writeln (weg.x);
          writeln (weg.y);
          writeln (weg.z);
          readln;
   end; {graphics}
(*   else
   Writeln('Graphics error:', GraphErrorMsg(ErrCode));
   Nc_reset;
*)
end. {program}
